from _nod import (
    open_disc_from_image,
    DiscBase,
    DiscBuilderGCN,
    Partition,
    PartReadStream,
    DolHeader,
    ExtractionContext,
    ProgressCallback,
)
from . import version

VERSION = version.version



from typing import Tuple, Optional

from cpython.bytes cimport PyBytes_FromStringAndSize
from libc.stddef cimport wchar_t
from libc.stdint cimport uint8_t, uint32_t, int64_t, uint64_t
from libcpp cimport bool as c_bool
from libcpp.string cimport string
from libcpp.memory cimport unique_ptr
from libcpp.functional cimport function


cdef extern from "string" namespace "std":
    cdef cppclass string_view:
        string_view()
        const char* data()
        size_t size() const 
        # wrap-ignore

    cdef cppclass optional[T]:
        c_bool operator bool()
        T& operator*()
        void operator=(T)
        T& value()


cdef extern from "nod/IDiscIO.hpp" namespace "nod":
    cppclass IReadStream:
        uint64_t read(void* buf, uint64_t length)
        void seek(int64_t offset, int whence)
        uint64_t position() const

    cppclass IPartReadStream(IReadStream):
        pass


cdef extern from "nod/DiscBase.hpp" namespace "nod":
    cdef cppclass EBuildResult:
        c_bool operator==(const EBuildResult&)

    cdef EBuildResult EBuildResult_Success "nod::EBuildResult::Success"
    cdef EBuildResult EBuildResult_Failed "nod::EBuildResult::Failed"
    cdef EBuildResult EBuildResult_DiskFull "nod::EBuildResult::DiskFull"
    
    cdef cppclass Kind:
        c_bool operator==(const Kind&)
        
    cdef Kind Kind_File "nod::Node::Kind::File"
    cdef Kind Kind_Directory "nod::Node::Kind::Directory"

    ctypedef function[void(float, string_view, size_t)] FProgress
    
    cppclass Node:

        cppclass DirectoryIterator:
            Node& operator*()
            c_bool operator==(const DirectoryIterator& other) const
            c_bool operator!=(const DirectoryIterator& other) const
            DirectoryIterator& operator++()

        unique_ptr[IPartReadStream] beginReadStream(uint64_t offset) const
        Kind getKind() const
        string_view getName() const
        DirectoryIterator find(string name) const
        DirectoryIterator begin()
        DirectoryIterator end()

    cppclass Header:
        char m_gameID[6]
        char m_discNum
        char m_discVersion
        char m_audioStreaming
        char m_streamBufSz
        char m_unk1[14]
        uint32_t m_wiiMagic
        uint32_t m_gcnMagic
        char m_gameTitle[64]
        char m_disableHashVerification
        char m_disableDiscEnc
        char m_unk2[0x39e]
        uint32_t m_debugMonOff
        uint32_t m_debugLoadAddr
        char m_unk3[0x18]
        uint32_t m_dolOff
        uint32_t m_fstOff
        uint32_t m_fstSz
        uint32_t m_fstMaxSz
        uint32_t m_fstMemoryAddress
        uint32_t m_userPosition
        uint32_t m_userSz
        uint8_t padding1[4]

    cppclass IPartition:
        c_bool extractToDirectory(string path, const ExtractionContext& ctx) except * const
        uint64_t getDOLSize() const
        const Header& getHeader() const
        const Node& getFSTRoot() const

    cdef cppclass DiscBase:
        IPartition* getDataPartition()
        IPartition* getUpdatePartition()


cdef extern  from "nod/DiscGCN.hpp" namespace "nod":
    cdef cppclass DiscBuilderGCN:
        DiscBuilderGCN(string outPath, FProgress progressCB)
        EBuildResult buildFromDirectory(string dirIn) except *

        @staticmethod
        optional[uint64_t] CalculateTotalSizeRequired(string dirIn)


cdef extern from "nod/nod.hpp" namespace "nod":
    cdef struct ExtractionContext:
        c_bool force
        function[void(string_view, float)] progressCB

    unique_ptr[DiscBase] OpenDiscFromImage(string path, c_bool& isWii)


cdef extern from "py-nod/nod_wrap_util.hpp" namespace "nod_wrap":
    function[void(string_view, float)] createProgressCallbackFunction(object, void (*)(object, const string&, float) except *)
    function[void(float, string_view, size_t)] createFProgressFunction(object, void (*)(object, float, const string&, size_t) except *)

    object getDol(const IPartition*)
    void doPrint(const IPartition*)
    void registerLogvisorToExceptionConverter()
    void removeLogvisorToExceptionConverter()
    object _handleNativeException(object)
    void checkException() except *



import hashlib
import os
from pathlib import Path

import pytest

import nod


def fprogress_callback(progress: float, name: str, bytes: int):
    pass


def sha256_checksum(filename, block_size=65536):
    sha256 = hashlib.sha256()
    with filename.open('rb') as f:
        for block in iter(lambda: f.read(block_size), b''):
            sha256.update(block)
    return sha256.hexdigest()


@pytest.fixture(name="pack_sample_game", scope="session")
def _pack_sample_game(tmp_path_factory):
    filesystem_root = Path(__file__).parent.joinpath("sample_game")
    image_out = tmp_path_factory.mktemp("iso").joinpath("game.iso")
    disc_builder = nod.DiscBuilderGCN(image_out, fprogress_callback)
    disc_builder.build_from_directory(filesystem_root)
    return image_out



def test_packunpack(pack_sample_game):
    iso_hash = sha256_checksum(pack_sample_game)
    assert iso_hash == '0d98f9bf2c94051bec6145373c0b2c4baee0b35ccb066eaf354fd5a7b9324816'


def test_build_and_check(pack_sample_game):
    result = nod.open_disc_from_image(pack_sample_game)
    assert result is not None
    assert not result[1]
    disc = result[0]

    files = disc.get_data_partition().files()
    assert sorted(files) == [
        "a.txt",
        "b.txt",
        "d/nested.txt",
    ]
    assert disc.get_data_partition().read_file("d/nested.txt").read(3) == b"abc"

    f = disc.get_data_partition().read_file("b.txt")
    f.seek(2, 1)
    assert f.read(1) == b"3"
    f.seek(1)
    assert f.read(2) == b"23"

    f.close()
    with pytest.raises(RuntimeError):
        f.read(1)



import os
from enum import Enum
from typing import Tuple, Optional, Callable, List
from contextlib import contextmanager

import cython
from cython.operator cimport dereference, preincrement
from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_AsString
from libcpp cimport bool as c_bool
from libcpp.string cimport string
from libcpp.memory cimport unique_ptr
from libcpp.utility cimport move

from nod_wrap cimport (
    optional as c_optional,
    string_view as c_string_view,

    ExtractionContext as c_ExtractionContext,
    createProgressCallbackFunction,
    getDol as _getDol,
    DiscBase as c_DiscBase,
    Header as c_Header,
    Kind_File,
    IPartReadStream as c_IPartReadStream,
    OpenDiscFromImage,
    DiscBuilderGCN as c_DiscBuilderGCN,
    createFProgressFunction,
    Node,
    EBuildResult,
    EBuildResult_Success,
    EBuildResult_Failed,
    EBuildResult_DiskFull,
    registerLogvisorToExceptionConverter,
    removeLogvisorToExceptionConverter,
    _handleNativeException,
    checkException,
)


cdef string _str_to_string(str path):
    return path.encode("utf-8")


cdef str _view_to_str(c_string_view str_view):
    return PyBytes_FromStringAndSize(str_view.data(), str_view.size()).decode("utf-8")


ProgressCallback = Callable[[float, str, int], None]


cdef void invoke_callback_function(object callback, const string& a, float progress) except *:
    callback(a.decode("utf-8"), progress)


cdef void invoke_fprogress_function(object callback, float totalProg, const string& fileName, size_t fileBytesXfered) except *:
    callback(totalProg, fileName.decode("utf-8"), fileBytesXfered)


@contextmanager
def _log_exception_handler():
    registerLogvisorToExceptionConverter()
    yield
    removeLogvisorToExceptionConverter()


cdef class DolHeader:
    @staticmethod
    cdef create(const c_Header& h):
        self = DolHeader()
        self.game_id = PyBytes_FromStringAndSize(h.m_gameID, 6)
        self.disc_num = h.m_discNum
        self.disc_version = h.m_discVersion
        self.audio_streaming = h.m_audioStreaming
        self.stream_buf_sz = h.m_streamBufSz
        self.wii_magic = h.m_wiiMagic
        self.gcn_magic = h.m_gcnMagic
        self.game_title = PyBytes_FromStringAndSize(h.m_gameTitle, 64)
        self.disable_hash_verification = h.m_disableHashVerification
        self.disable_disc_enc = h.m_disableDiscEnc
        self.debug_mon_off = h.m_debugMonOff
        self.debug_load_addr = h.m_debugLoadAddr
        self.dol_off = h.m_dolOff
        self.fst_off = h.m_fstOff
        self.fst_sz = h.m_fstSz
        self.fst_max_sz = h.m_fstMaxSz
        self.fst_memory_address = h.m_fstMemoryAddress
        self.user_position = h.m_userPosition
        self.user_sz = h.m_userSz


cdef class ExtractionContext:
    cdef c_ExtractionContext c_context

    def __cinit__(self):
        self.c_context = c_ExtractionContext()

    @property
    def force(self):
        return self.c_context.force

    @force.setter
    def force(self, value):
        self.c_context.force = value

    def set_progress_callback(self, callback):
        self.c_context.progressCB = createProgressCallbackFunction(callback, invoke_callback_function)


cdef class PartReadStream:
    cdef unique_ptr[c_IPartReadStream] c_stream
    cdef int offset

    @staticmethod
    cdef create(unique_ptr[c_IPartReadStream] c_stream):
        stream = PartReadStream()
        stream.c_stream = move(c_stream)
        stream.offset = stream.c_stream.get().position()
        return stream

    def read(self, length):
        if not self.c_stream:
            raise RuntimeError("already closed")
        buf = PyBytes_FromStringAndSize(NULL, length)
        self.c_stream.get().read(PyBytes_AsString(buf), length)
        return buf

    def seek(self, offset, whence=0):
        if not self.c_stream:
            raise RuntimeError("already closed")
        if whence == 0:
            offset += self.offset
        self.c_stream.get().seek(offset, whence)
    
    def tell(self):
        if not self.c_stream:
            raise RuntimeError("already closed")
        return self.c_stream.get().position() - self.offset

    def close(self):
        self.c_stream.reset()


cdef _files_for(Node& node, prefix: str, result: list):
    cdef c_optional[Node.DirectoryIterator] f

    name = _view_to_str(node.getName())
    if node.getKind() == Kind_File:
        result.append(prefix + name)
    else:
        newPrefix = prefix
        if name:
            newPrefix = prefix + name + "/"
        f = node.begin()
        while f.value() != node.end():
            _files_for(dereference(dereference(f)), newPrefix, result)
            preincrement(dereference(f))


cdef class Partition:
    cdef c_DiscBase.IPartition* c_partition
    cdef object discParent

    def __init__(self, parent):
        self.discParent = parent

    @staticmethod
    cdef create(c_DiscBase.IPartition* c_partition, object parent):
        partition = Partition(parent)
        partition.c_partition = c_partition
        return partition

    def get_dol(self) -> bytes:
        return _getDol(self.c_partition)

    def get_header(self):
        return DolHeader.create(self.c_partition.getHeader())

    def extract_to_directory(self, path: str, context: ExtractionContext) -> None:
        def work():
            with _log_exception_handler():
                extraction_successful = self.c_partition.extractToDirectory(
                    _str_to_string(path),
                    context.c_context
                )
            if not extraction_successful:
                raise RuntimeError("Unable to extract")
        return _handleNativeException(work)

    def files(self) -> List[str]:
        cdef Node* node = &self.c_partition.getFSTRoot()
        result = []
        _files_for(dereference(node), "", result)
        return result


    def read_file(self, path: str, offset: int = 0) -> PartReadStream:
        cdef Node* node = &self.c_partition.getFSTRoot()
        cdef c_optional[Node.DirectoryIterator] f

        for part in path.split("/"):
            f = node.find(_str_to_string(part))
            if f.value() != node.end():
                node = &dereference(dereference(f))
            else:
                raise Exception(f"File {part} not found in '{_view_to_str(node.getName())}'")
            
        return PartReadStream.create(dereference(dereference(f)).beginReadStream(offset))


cdef class DiscBase:
    cdef unique_ptr[c_DiscBase] c_disc

    def get_data_partition(self) -> Optional[Partition]:
        cdef c_DiscBase.IPartition*partition = self.c_disc.get().getDataPartition()
        if partition:
            return Partition.create(partition, self)
        else:
            return None


cdef class DiscBuilderGCN:
    cdef c_DiscBuilderGCN* c_builder

    def __init__(self, out_path: os.PathLike, progress_callback: ProgressCallback):
        pass

    def __cinit__(self, out_path: os.PathLike, progress_callback: ProgressCallback):
        self.c_builder = new c_DiscBuilderGCN(_str_to_string(os.fspath(out_path)),
                                              createFProgressFunction(progress_callback, invoke_fprogress_function))

    def __dealloc__(self):
        del self.c_builder

    def build_from_directory(self, directory_in: os.PathLike) -> None:
        def work():
            with _log_exception_handler():
                self.c_builder.buildFromDirectory(_str_to_string(os.fspath(directory_in)))
        return _handleNativeException(work)

    @staticmethod
    def calculate_total_size_required(directory_in: os.PathLike) -> Optional[int]:
        size = c_DiscBuilderGCN.CalculateTotalSizeRequired(_str_to_string(os.fspath(directory_in)))
        if size:
            return cython.operator.dereference(size)

        return None


def open_disc_from_image(path: os.PathLike) -> Optional[Tuple[DiscBase, bool]]:
    def work():
        disc = DiscBase()
        cdef c_bool is_wii = True

        with _log_exception_handler():
            disc.c_disc = OpenDiscFromImage(_str_to_string(os.fspath(path)), is_wii)
            checkException()
            return disc, is_wii

    return _handleNativeException(work)
